@@ -69,8 +69,8 @@
 some syntax analysis.")
 
 (defvar go-mode-font-lock-keywords
-  (let ((builtins '("cap" "close" "closed" "len" "make" "new"
-                    "panic" "panicln" "print" "println"))
+  (let ((builtins '("append" "cap" "close" "complex" "copy" "imag" "len"
+                    "make" "new" "panic" "print" "println" "real" "recover"))
         (constants '("nil" "true" "false" "iota"))
         (type-name "\\s *\\(?:[*(]\\s *\\)*\\(?:\\w+\\s *\\.\\s *\\)?\\(\\w+\\)")
         )
@@ -507,7 +507,9 @@ Replace the current buffer on success; display errors on failure."
  (let ((srcbuf (current-buffer)))
    (with-temp-buffer
      (let ((outbuf (current-buffer))
-           (errbuf (get-buffer-create "*Gofmt Errors*")))
+           (errbuf (get-buffer-create "*Gofmt Errors*"))
+           (coding-system-for-read 'utf-8)    ;; use utf-8 with subprocesses
+           (coding-system-for-write 'utf-8))
        (with-current-buffer errbuf (erase-buffer))
        (with-current-buffer srcbuf
          (save-restriction
@@ -521,7 +523,7 @@ Replace the current buffer on success; display errors on failure."
                    (erase-buffer)
                    (insert-buffer-substring outbuf)
                    (goto-char (min old-point (point-max)))
-                   (if old-mark (set-mark (min old-mark (point-max))))
+                   (if old-mark (push-mark (min old-mark (point-max)) t))
                    (kill-buffer errbuf))
 
                ;; gofmt failed: display the errors
