@@ -7,107 +7,9 @@
 #include "gg.h"
 #include "opt.h"
 
-static Prog *pret;
-
 void
-compile(Node *fn)
+defframe(Prog *ptxt)
 {
-	Plist *pl;
-	Node nod1;
-	Prog *ptxt;
-	int32 lno;
-	Type *t;
-	Iter save;
-
-	if(newproc == N) {
-		newproc = sysfunc("newproc");
-		deferproc = sysfunc("deferproc");
-		deferreturn = sysfunc("deferreturn");
-		panicindex = sysfunc("panicindex");
-		panicslice = sysfunc("panicslice");
-		throwreturn = sysfunc("throwreturn");
-	}
-
-	if(fn->nbody == nil)
-		return;
-
-	// set up domain for labels
-	clearlabels();
-
-	lno = setlineno(fn);
-
-	curfn = fn;
-	dowidth(curfn->type);
-
-	if(curfn->type->outnamed) {
-		// add clearing of the output parameters
-		t = structfirst(&save, getoutarg(curfn->type));
-		while(t != T) {
-			if(t->nname != N)
-				curfn->nbody = concat(list1(nod(OAS, t->nname, N)), curfn->nbody);
-			t = structnext(&save);
-		}
-	}
-
-	hasdefer = 0;
-	walk(curfn);
-	if(nerrors != 0 || isblank(curfn->nname))
-		goto ret;
-
-	allocparams();
-
-	continpc = P;
-	breakpc = P;
-
-	pl = newplist();
-	pl->name = curfn->nname;
-	
-	setlineno(curfn);
-
-	nodconst(&nod1, types[TINT32], 0);
-	ptxt = gins(ATEXT, curfn->nname, &nod1);
-	afunclit(&ptxt->from);
-
-	ginit();
-	genlist(curfn->enter);
-	
-	pret = nil;
-	if(hasdefer || curfn->exit) {
-		Prog *p1;
-
-		p1 = gjmp(nil);
-		pret = gjmp(nil);
-		patch(p1, pc);
-	}
-
-	genlist(curfn->nbody);
-	gclean();
-	checklabels();
-	if(nerrors != 0)
-		goto ret;
-
-	if(curfn->type->outtuple != 0)
-		ginscall(throwreturn, 0);
-
-	if(pret)
-		patch(pret, pc);
-	ginit();
-	if(hasdefer)
-		ginscall(deferreturn, 0);
-	if(curfn->exit)
-		genlist(curfn->exit);
-	gclean();
-	if(nerrors != 0)
-		goto ret;
-	if(curfn->endlineno)
-		lineno = curfn->endlineno;
-	pc->as = ARET;	// overwrite AEND
-	pc->lineno = lineno;
-
-	if(!debug['N'] || debug['R'] || debug['P']) {
-		regopt(ptxt);
-	}
-
 	// fill in argument size
 	ptxt->to.type = D_CONST2;
 	ptxt->reg = 0; // flags
@@ -118,12 +20,32 @@ compile(Node *fn)
 		maxstksize = stksize;
 	ptxt->to.offset = rnd(maxstksize+maxarg, widthptr);
 	maxstksize = 0;
+}
 
-	if(debug['f'])
-		frame(0);
+// Sweep the prog list to mark any used nodes.
+void
+markautoused(Prog* p)
+{
+	for (; p; p = p->link) {
+		if (p->from.name == D_AUTO && p->from.node)
+			p->from.node->used++;
 
-ret:
-	lineno = lno;
+		if (p->to.name == D_AUTO && p->to.node)
+			p->to.node->used++;
+	}
+}
+
+// Fixup instructions after compactframe has moved all autos around.
+void
+fixautoused(Prog* p)
+{
+	for (; p; p = p->link) {
+		if (p->from.name == D_AUTO && p->from.node)
+			p->from.offset += p->from.node->stkdelta;
+
+		if (p->to.name == D_AUTO && p->to.node)
+			p->to.offset += p->to.node->stkdelta;
+	}
 }
 
 /*
@@ -197,7 +119,7 @@ ginscall(Node *f, int proc)
 			nodconst(&con, types[TINT32], 0);
 			p = gins(ACMP, &con, N);
 			p->reg = 0;
-			patch(gbranch(ABNE, T), pret);
+			patch(gbranch(ABNE, T), retpc);
 		}
 		break;
 	}
@@ -251,13 +173,13 @@ cgen_callinter(Node *n, Node *res, int proc)
 	nodindreg(&nodsp, types[tptr], REGSP);
 	nodsp.xoffset = 4;
 	nodo.xoffset += widthptr;
-	cgen(&nodo, &nodsp);	// 4(SP) = 8(REG) -- i.s
+	cgen(&nodo, &nodsp);	// 4(SP) = 4(REG) -- i.data
 
 	nodo.xoffset -= widthptr;
-	cgen(&nodo, &nodr);	// REG = 0(REG) -- i.m
+	cgen(&nodo, &nodr);	// REG = 0(REG) -- i.tab
 
 	nodo.xoffset = n->left->xoffset + 3*widthptr + 8;
-	cgen(&nodo, &nodr);	// REG = 32+offset(REG) -- i.m->fun[f]
+	cgen(&nodo, &nodr);	// REG = 20+offset(REG) -- i.tab->fun[f]
 
 	// BOTCH nodr.type = fntype;
 	nodr.type = n->left->type;
@@ -401,7 +323,7 @@ cgen_ret(Node *n)
 {
 	genlist(n->list);		// copy out args
 	if(hasdefer || curfn->exit)
-		gjmp(pret);
+		gjmp(retpc);
 	else
 		gins(ARET, N, N);
 }
@@ -548,9 +470,10 @@ samereg(Node *a, Node *b)
 void
 cgen_shift(int op, Node *nl, Node *nr, Node *res)
 {
-	Node n1, n2, n3, t;
+	Node n1, n2, n3, nt, t, lo, hi;
 	int w;
 	Prog *p1, *p2, *p3;
+	Type *tr;
 	uvlong sc;
 
 	if(nl->type->width > 4)
@@ -582,16 +505,42 @@ cgen_shift(int op, Node *nl, Node *nr, Node *res)
 		return;
 	}
 
-	if(nl->ullman >= nr->ullman) {
-		regalloc(&n2, nl->type, res);
-		cgen(nl, &n2);
-		regalloc(&n1, nr->type, N);
-		cgen(nr, &n1);
+	tr = nr->type;
+	if(tr->width > 4) {
+		tempname(&nt, nr->type);
+		if(nl->ullman >= nr->ullman) {
+			regalloc(&n2, nl->type, res);
+			cgen(nl, &n2);
+			cgen(nr, &nt);
+			n1 = nt;
+		} else {
+			cgen(nr, &nt);
+			regalloc(&n2, nl->type, res);
+			cgen(nl, &n2);
+		}
+		split64(&nt, &lo, &hi);
+		regalloc(&n1, types[TUINT32], N);
+		regalloc(&n3, types[TUINT32], N);
+		gmove(&lo, &n1);
+		gmove(&hi, &n3);
+		gins(ATST, &n3, N);
+		nodconst(&t, types[TUINT32], w);
+		p1 = gins(AMOVW, &t, &n1);
+		p1->scond = C_SCOND_NE;
+		tr = types[TUINT32];
+		regfree(&n3);
 	} else {
-		regalloc(&n1, nr->type, N);
-		cgen(nr, &n1);
-		regalloc(&n2, nl->type, res);
-		cgen(nl, &n2);
+		if(nl->ullman >= nr->ullman) {
+			regalloc(&n2, nl->type, res);
+			cgen(nl, &n2);
+			regalloc(&n1, nr->type, N);
+			cgen(nr, &n1);
+		} else {
+			regalloc(&n1, nr->type, N);
+			cgen(nr, &n1);
+			regalloc(&n2, nl->type, res);
+			cgen(nl, &n2);
+		}
 	}
 
 	// test for shift being 0
@@ -599,7 +548,7 @@ cgen_shift(int op, Node *nl, Node *nr, Node *res)
 	p3 = gbranch(ABEQ, T);
 
 	// test and fix up large shifts
-	regalloc(&n3, nr->type, N);
+	regalloc(&n3, tr, N);
 	nodconst(&t, types[TUINT32], w);
 	gmove(&t, &n3);
 	gcmp(ACMP, &n1, &n3);
