@@ -930,9 +930,6 @@ cgen(Node *n, Node *nn)
 
 			return;
 		}
-		o = 0;
-		if(REGARG >= 0)
-			o = reg[REGARG];
 		gargs(r, &nod, &nod1);
 		if(l->addable < INDEXED) {
 			reglcgen(&nod, l, nn);
@@ -941,9 +938,8 @@ cgen(Node *n, Node *nn)
 			regfree(&nod);
 		} else
 			gopcode(OFUNC, n->type, Z, l);
-		if(REGARG >= 0)
-			if(o != reg[REGARG])
-				reg[REGARG]--;
+		if(REGARG >= 0 && reg[REGARG])
+			reg[REGARG]--;
 		if(nn != Z) {
 			regret(&nod, n);
 			gmove(&nod, nn);
@@ -1544,7 +1540,7 @@ sugen(Node *n, Node *nn, int32 w)
 			nod0.addable = 0;
 			nod0.right = l;
 
-			/* prtree(&nod0, "hand craft"); /* */
+			// prtree(&nod0, "hand craft");
 			cgen(&nod0, Z);
 		}
 		break;
