@@ -36,6 +36,7 @@
 #include	"../ld/pe.h"
 
 void	dynreloc(void);
+static vlong addaddrplus4(Sym *s, Sym *t, int32 add);
 
 /*
  * divide-and-conquer list-link
@@ -249,17 +250,25 @@ dynrelocsym(Sym *s)
 			return;
 		for(r=s->r; r<s->r+s->nr; r++) {
 			targ = r->sym;
-			if(r->sym->plt == -2) { // make dynimport JMP table for PE object files.
+			if(r->sym->plt == -2 && r->sym->got != -2) { // make dynimport JMP table for PE object files.
 				targ->plt = rel->size;
 				r->sym = rel;
 				r->add = targ->plt;
 				
 				// jmp *addr
-				adduint8(rel, 0xff);
-				adduint8(rel, 0x25);
-				addaddr(rel, targ);
-				adduint8(rel, 0x90);
-				adduint8(rel, 0x90);
+				if(thechar == '8') {
+					adduint8(rel, 0xff);
+					adduint8(rel, 0x25);
+					addaddr(rel, targ);
+					adduint8(rel, 0x90);
+					adduint8(rel, 0x90);
+				} else {
+					adduint8(rel, 0xff);
+					adduint8(rel, 0x24);
+					adduint8(rel, 0x25);
+					addaddrplus4(rel, targ, 0);
+					adduint8(rel, 0x90);
+				}
 			} else if(r->sym->plt >= 0) {
 				r->sym = rel;
 				r->add = targ->plt;
@@ -278,6 +287,10 @@ dynreloc(void)
 {
 	Sym *s;
 	
+	// -d supresses dynamic loader format, so we may as well not
+	// compute these sections or mark their symbols as reachable.
+	if(debug['d'] && HEADTYPE != Hwindows)
+		return;
 	if(debug['v'])
 		Bprint(&bso, "%5.2f reloc\n", cputime());
 	Bflush(&bso);
@@ -447,7 +460,7 @@ codeblk(int32 addr, int32 size)
 	uchar *q;
 
 	if(debug['a'])
-		Bprint(&bso, "codeblk [%#x,%#x) at offset %#llx\n", addr, addr+size, seek(cout, 0, 1));
+		Bprint(&bso, "codeblk [%#x,%#x) at offset %#llx\n", addr, addr+size, cpos());
 
 	blk(textp, addr, size);
 
@@ -482,13 +495,13 @@ codeblk(int32 addr, int32 size)
 			q = sym->p;
 			
 			while(n >= 16) {
-				Bprint(&bso, "%.6ux\t%-20.16I\n",  addr, q);
+				Bprint(&bso, "%.6ux\t%-20.16I\n", addr, q);
 				addr += 16;
 				q += 16;
 				n -= 16;
 			}
 			if(n > 0)
-				Bprint(&bso, "%.6ux\t%-20.*I\n", addr, n, q);
+				Bprint(&bso, "%.6ux\t%-20.*I\n", addr, (int)n, q);
 			addr += n;
 			continue;
 		}
@@ -499,10 +512,10 @@ codeblk(int32 addr, int32 size)
 				epc = p->link->pc;
 			else
 				epc = sym->value + sym->size;
-			Bprint(&bso, "%.6ux\t", p->pc);
+			Bprint(&bso, "%.6llux\t", (uvlong)p->pc);
 			q = sym->p + p->pc - sym->value;
 			n = epc - p->pc;
-			Bprint(&bso, "%-20.*I | %P\n", n, q, p);
+			Bprint(&bso, "%-20.*I | %P\n", (int)n, q, p);
 			addr += n;
 		}
 	}
@@ -523,7 +536,7 @@ datblk(int32 addr, int32 size)
 	uchar *p, *ep;
 
 	if(debug['a'])
-		Bprint(&bso, "datblk [%#x,%#x) at offset %#llx\n", addr, addr+size, seek(cout, 0, 1));
+		Bprint(&bso, "datblk [%#x,%#x) at offset %#llx\n", addr, addr+size, cpos());
 
 	blk(datap, addr, size);
 
@@ -543,7 +556,7 @@ datblk(int32 addr, int32 size)
 			Bprint(&bso, "%-20s %.8ux| 00 ...\n", "(pre-pad)", addr);
 			addr = sym->value;
 		}
-		Bprint(&bso, "%-20s %.8ux|", sym->name, addr);
+		Bprint(&bso, "%-20s %.8ux|", sym->name, (uint)addr);
 		p = sym->p;
 		ep = p + sym->np;
 		while(p < ep)
@@ -555,8 +568,8 @@ datblk(int32 addr, int32 size)
 	}
 
 	if(addr < eaddr)
-		Bprint(&bso, "%-20s %.8ux| 00 ...\n", "(post-pad)", addr);
-	Bprint(&bso, "%-20s %.8ux|\n", "", eaddr);
+		Bprint(&bso, "%-20s %.8ux| 00 ...\n", "(post-pad)", (uint)addr);
+	Bprint(&bso, "%-20s %.8ux|\n", "", (uint)eaddr);
 }
 
 void
@@ -674,6 +687,27 @@ addaddrplus(Sym *s, Sym *t, int32 add)
 	return i;
 }
 
+static vlong
+addaddrplus4(Sym *s, Sym *t, int32 add)
+{
+	vlong i;
+	Reloc *r;
+
+	if(s->type == 0)
+		s->type = SDATA;
+	s->reachable = 1;
+	i = s->size;
+	s->size += 4;
+	symgrow(s, s->size);
+	r = addrel(s);
+	r->sym = t;
+	r->off = i;
+	r->siz = 4;
+	r->type = D_ADDR;
+	r->add = add;
+	return i;
+}
+
 vlong
 addpcrelplus(Sym *s, Sym *t, int32 add)
 {
@@ -781,22 +815,63 @@ dodata(void)
 	 */
 
 	/* read-only data */
-	sect = addsection(&segtext, ".rodata", 06);
+	sect = addsection(&segtext, ".rodata", 04);
 	sect->vaddr = 0;
 	datsize = 0;
 	s = datap;
-	for(; s != nil && s->type < SDATA; s = s->next) {
+	for(; s != nil && s->type < SSYMTAB; s = s->next) {
 		s->type = SRODATA;
-		t = rnd(s->size, PtrSize);
 		s->value = datsize;
-		datsize += t;
+		datsize += rnd(s->size, PtrSize);
 	}
 	sect->len = datsize - sect->vaddr;
-	
-	/* data */
+
+	/* gosymtab */
+	sect = addsection(&segtext, ".gosymtab", 04);
+	sect->vaddr = datsize;
+	for(; s != nil && s->type < SPCLNTAB; s = s->next) {
+		s->type = SRODATA;
+		s->value = datsize;
+		datsize += s->size;
+	}
+	sect->len = datsize - sect->vaddr;
+	datsize = rnd(datsize, PtrSize);
+
+	/* gopclntab */
+	sect = addsection(&segtext, ".gopclntab", 04);
+	sect->vaddr = datsize;
+	for(; s != nil && s->type < SELFROSECT; s = s->next) {
+		s->type = SRODATA;
+		s->value = datsize;
+		datsize += s->size;
+	}
+	sect->len = datsize - sect->vaddr;
+	datsize = rnd(datsize, PtrSize);
+
+	/* read-only ELF sections */
+	for(; s != nil && s->type < SELFSECT; s = s->next) {
+		sect = addsection(&segtext, s->name, 04);
+		sect->vaddr = datsize;
+		s->type = SRODATA;
+		s->value = datsize;
+		datsize += rnd(s->size, PtrSize);
+		sect->len = datsize - sect->vaddr;
+	}
+
+	/* writable ELF sections */
 	datsize = 0;
+	for(; s != nil && s->type < SDATA; s = s->next) {
+		sect = addsection(&segdata, s->name, 06);
+		sect->vaddr = datsize;
+		s->type = SDATA;
+		s->value = datsize;
+		datsize += rnd(s->size, PtrSize);
+		sect->len = datsize - sect->vaddr;
+	}
+
+	/* data */
 	sect = addsection(&segdata, ".data", 06);
-	sect->vaddr = 0;
+	sect->vaddr = datsize;
 	for(; s != nil && s->type < SBSS; s = s->next) {
 		s->type = SDATA;
 		t = s->size;
@@ -804,9 +879,13 @@ dodata(void)
 			diag("%s: no size", s->name);
 			t = 1;
 		}
-		if(t & 1)
+		if(t >= PtrSize)
+			t = rnd(t, PtrSize);
+		else if(t > 2)
+			t = rnd(t, 4);
+		if(t & 1) {
 			;
-		else if(t & 2)
+		} else if(t & 2)
 			datsize = rnd(datsize, 2);
 		else if(t & 4)
 			datsize = rnd(datsize, 4);
@@ -826,9 +905,13 @@ dodata(void)
 			diag("unexpected symbol type %d", s->type);
 		}
 		t = s->size;
-		if(t & 1)
+		if(t >= PtrSize)
+			t = rnd(t, PtrSize);
+		else if(t > 2)
+			t = rnd(t, 4);
+		if(t & 1) {
 			;
-		else if(t & 2)
+		} else if(t & 2)
 			datsize = rnd(datsize, 2);
 		else if(t & 4)
 			datsize = rnd(datsize, 4);
@@ -878,7 +961,7 @@ textaddress(void)
 void
 address(void)
 {
-	Section *s, *text, *data, *rodata;
+	Section *s, *text, *data, *rodata, *symtab, *pclntab;
 	Sym *sym, *sub;
 	uvlong va;
 
@@ -888,38 +971,43 @@ address(void)
 	segtext.fileoff = HEADR;
 	for(s=segtext.sect; s != nil; s=s->next) {
 		s->vaddr = va;
-		va += s->len;
-		segtext.len = va - INITTEXT;
-		va = rnd(va, INITRND);
+		va += rnd(s->len, PtrSize);
 	}
+	segtext.len = va - INITTEXT;
 	segtext.filelen = segtext.len;
 
+	va = rnd(va, INITRND);
+
 	segdata.rwx = 06;
 	segdata.vaddr = va;
 	segdata.fileoff = va - segtext.vaddr + segtext.fileoff;
+	segdata.filelen = 0;
 	if(HEADTYPE == Hwindows)
 		segdata.fileoff = segtext.fileoff + rnd(segtext.len, PEFILEALIGN);
-	if(HEADTYPE == Hplan9x32) {
-		segdata.vaddr = va = rnd(va, 4096);
+	if(HEADTYPE == Hplan9x32)
 		segdata.fileoff = segtext.fileoff + segtext.filelen;
-	}
+	data = nil;
 	for(s=segdata.sect; s != nil; s=s->next) {
 		s->vaddr = va;
 		va += s->len;
+		segdata.filelen += s->len;
 		segdata.len = va - segdata.vaddr;
+		if(strcmp(s->name, ".data") == 0)
+			data = s;
 	}
-	segdata.filelen = segdata.sect->len;	// assume .data is first
-	
+	segdata.filelen -= data->next->len; // deduct .bss
+
 	text = segtext.sect;
-	rodata = segtext.sect->next;
-	data = segdata.sect;
+	rodata = text->next;
+	symtab = rodata->next;
+	pclntab = symtab->next;
 
 	for(sym = datap; sym != nil; sym = sym->next) {
 		cursym = sym;
 		if(sym->type < SDATA)
 			sym->value += rodata->vaddr;
 		else
-			sym->value += data->vaddr;
+			sym->value += segdata.sect->vaddr;
 		for(sub = sym->sub; sub != nil; sub = sub->sub)
 			sub->value += sym->value;
 	}
@@ -928,12 +1016,11 @@ address(void)
 	xdefine("etext", STEXT, text->vaddr + text->len);
 	xdefine("rodata", SRODATA, rodata->vaddr);
 	xdefine("erodata", SRODATA, rodata->vaddr + rodata->len);
+	xdefine("symtab", SRODATA, symtab->vaddr);
+	xdefine("esymtab", SRODATA, symtab->vaddr + symtab->len);
+	xdefine("pclntab", SRODATA, pclntab->vaddr);
+	xdefine("epclntab", SRODATA, pclntab->vaddr + pclntab->len);
 	xdefine("data", SBSS, data->vaddr);
 	xdefine("edata", SBSS, data->vaddr + data->len);
 	xdefine("end", SBSS, segdata.vaddr + segdata.len);
-
-	sym = lookup("pclntab", 0);
-	xdefine("epclntab", SRODATA, sym->value + sym->size);
-	sym = lookup("symtab", 0);
-	xdefine("esymtab", SRODATA, sym->value + sym->size);
 }
