@@ -529,8 +529,10 @@ find_or_diag(DWDie *die, char* name)
 {
 	DWDie *r;
 	r = find(die, name);
-	if (r == nil)
+	if (r == nil) {
 		diag("dwarf find: %s has no %s", getattr(die, DW_AT_name)->data, name);
+		errorexit();
+	}
 	return r;
 }
 
@@ -613,7 +615,7 @@ putattr(int form, int cls, vlong value, char *data)
 
 	case DW_FORM_ref_addr:	// reference to a DIE in the .info section
 		if (data == nil) {
-			diag("null dwarf reference");
+			diag("dwarf: null reference");
 			LPUT(0);  // invalid dwarf, gdb will complain.
 		} else {
 			if (((DWDie*)data)->offs == 0)
@@ -631,7 +633,7 @@ putattr(int form, int cls, vlong value, char *data)
 	case DW_FORM_strp:	// string
 	case DW_FORM_indirect:	// (see Section 7.5.3)
 	default:
-		diag("Unsupported atribute form %d / class %d", form, cls);
+		diag("dwarf: unsupported attribute form %d / class %d", form, cls);
 		errorexit();
 	}
 }
@@ -823,7 +825,7 @@ decode_inuxi(uchar* p, int sz)
 		inuxi = inuxi8;
 		break;
 	default:
-		diag("decode inuxi %d", sz);
+		diag("dwarf: decode inuxi %d", sz);
 		errorexit();
 	}
 	for (i = 0; i < sz; i++)
@@ -1013,7 +1015,7 @@ defgotype(Sym *gotype)
 		return find_or_diag(&dwtypes, "<unspecified>");
 
 	if (strncmp("type.", gotype->name, 5) != 0) {
-		diag("Type name doesn't start with \".type\": %s", gotype->name);
+		diag("dwarf: type name doesn't start with \".type\": %s", gotype->name);
 		return find_or_diag(&dwtypes, "<unspecified>");
 	}
 	name = gotype->name + 5;  // could also decode from Type.string
@@ -1164,7 +1166,7 @@ defgotype(Sym *gotype)
 		break;
 
 	default:
-		diag("definition of unknown kind %d: %s", kind, gotype->name);
+		diag("dwarf: definition of unknown kind %d: %s", kind, gotype->name);
 		die = newdie(&dwtypes, DW_ABRV_TYPEDECL, name);
 		newrefattr(die, DW_AT_type, find_or_diag(&dwtypes, "<unspecified>"));
 	 }
@@ -1207,7 +1209,7 @@ copychildren(DWDie *dst, DWDie *src)
 }
 
 // Search children (assumed to have DW_TAG_member) for the one named
-// field and set it's DW_AT_type to dwtype
+// field and set its DW_AT_type to dwtype
 static void
 substitutetype(DWDie *structdie, char *field, DWDie* dwtype)
 {
@@ -1346,7 +1348,7 @@ synthesizemaptypes(DWDie *die)
 			valtype = defptrto(valtype);
 		newrefattr(fld, DW_AT_type, valtype);
 		newmemberoffsetattr(fld, hashsize + datavo);
-		newattr(dwhe, DW_AT_byte_size, DW_CLS_CONSTANT, hashsize + datsize, NULL);
+		newattr(dwhe, DW_AT_byte_size, DW_CLS_CONSTANT, hashsize + datsize, nil);
 
 		// Construct hash_subtable<hash_entry<K,V>>
 		dwhs = newdie(&dwtypes, DW_ABRV_STRUCTTYPE,
@@ -1357,7 +1359,7 @@ synthesizemaptypes(DWDie *die)
 		substitutetype(dwhs, "end", defptrto(dwhe));
 		substitutetype(dwhs, "entry", dwhe);  // todo: []hash_entry with dynamic size
 		newattr(dwhs, DW_AT_byte_size, DW_CLS_CONSTANT,
-			getattr(hash_subtable, DW_AT_byte_size)->value, NULL);
+			getattr(hash_subtable, DW_AT_byte_size)->value, nil);
 
 		// Construct hash<K,V>
 		dwh = newdie(&dwtypes, DW_ABRV_STRUCTTYPE,
@@ -1367,7 +1369,7 @@ synthesizemaptypes(DWDie *die)
 		copychildren(dwh, hash);
 		substitutetype(dwh, "st", defptrto(dwhs));
 		newattr(dwh, DW_AT_byte_size, DW_CLS_CONSTANT,
-			getattr(hash, DW_AT_byte_size)->value, NULL);
+			getattr(hash, DW_AT_byte_size)->value, nil);
 
 		newrefattr(die, DW_AT_type, defptrto(dwh));
 	}
@@ -1399,30 +1401,29 @@ synthesizechantypes(DWDie *die)
 		// sudog<T>
 		dws = newdie(&dwtypes, DW_ABRV_STRUCTTYPE,
 			mkinternaltypename("sudog",
-				getattr(elemtype, DW_AT_name)->data, NULL));
+				getattr(elemtype, DW_AT_name)->data, nil));
 		copychildren(dws, sudog);
 		substitutetype(dws, "elem", elemtype);
 		newattr(dws, DW_AT_byte_size, DW_CLS_CONSTANT,
-			sudogsize + (elemsize > 8 ? elemsize - 8 : 0), NULL);
+			sudogsize + (elemsize > 8 ? elemsize - 8 : 0), nil);
 
 		// waitq<T>
 		dww = newdie(&dwtypes, DW_ABRV_STRUCTTYPE,
-			mkinternaltypename("waitq", getattr(elemtype, DW_AT_name)->data, NULL));
+			mkinternaltypename("waitq", getattr(elemtype, DW_AT_name)->data, nil));
 		copychildren(dww, waitq);
 		substitutetype(dww, "first", defptrto(dws));
 		substitutetype(dww, "last",  defptrto(dws));
 		newattr(dww, DW_AT_byte_size, DW_CLS_CONSTANT,
-			getattr(waitq, DW_AT_byte_size)->value, NULL);
+			getattr(waitq, DW_AT_byte_size)->value, nil);
 
 		// hchan<T>
 		dwh = newdie(&dwtypes, DW_ABRV_STRUCTTYPE,
-			mkinternaltypename("hchan", getattr(elemtype, DW_AT_name)->data, NULL));
+			mkinternaltypename("hchan", getattr(elemtype, DW_AT_name)->data, nil));
 		copychildren(dwh, hchan);
 		substitutetype(dwh, "recvq", dww);
 		substitutetype(dwh, "sendq", dww);
-		substitutetype(dwh, "free", defptrto(dws));
 		newattr(dwh, DW_AT_byte_size, DW_CLS_CONSTANT,
-			getattr(hchan, DW_AT_byte_size)->value, NULL);
+			getattr(hchan, DW_AT_byte_size)->value, nil);
 
 		newrefattr(die, DW_AT_type, defptrto(dwh));
 	}
@@ -1434,6 +1435,7 @@ defdwsymb(Sym* sym, char *s, int t, vlong v, vlong size, int ver, Sym *gotype)
 {
 	DWDie *dv, *dt;
 
+	USED(size);
 	if (strncmp(s, "go.string.", 10) == 0)
 		return;
 
@@ -1513,12 +1515,12 @@ decodez(char *s)
 	ss = s + 1;	// first is 0
 	while((o = ((uint8)ss[0] << 8) | (uint8)ss[1]) != 0) {
 		if (o < 0 || o >= ftabsize) {
-			diag("corrupt z entry");
+			diag("dwarf: corrupt z entry");
 			return 0;
 		}
 		f = ftab[o];
 		if (f == nil) {
-			diag("corrupt z entry");
+			diag("dwarf: corrupt z entry");
 			return 0;
 		}
 		len += strlen(f) + 1;	// for the '/'
@@ -1576,13 +1578,16 @@ addhistfile(char *zentry)
 		histfile[histfilesize++] = "<eof>";
 
 	fname = decodez(zentry);
+//	print("addhistfile %d: %s\n", histfilesize, fname);
 	if (fname == 0)
 		return -1;
+
 	// Don't fill with duplicates (check only top one).
 	if (strcmp(fname, histfile[histfilesize-1]) == 0) {
 		free(fname);
 		return histfilesize - 1;
 	}
+
 	histfile[histfilesize++] = fname;
 	return histfilesize - 1;
 }
@@ -1590,7 +1595,7 @@ addhistfile(char *zentry)
 // if the histfile stack contains ..../runtime/runtime_defs.go
 // use that to set gdbscript
 static void
-finddebugruntimepath()
+finddebugruntimepath(void)
 {
 	int i, l;
 	char *c;
@@ -1606,11 +1611,13 @@ finddebugruntimepath()
 }
 
 // Go's runtime C sources are sane, and Go sources nest only 1 level,
-// so 16 should be plenty.
+// so a handful would be plenty, if it weren't for the fact that line
+// directives can push an unlimited number of them.
 static struct {
 	int file;
 	vlong line;
-} includestack[16];
+} *includestack;
+static int includestacksize;
 static int includetop;
 static vlong absline;
 
@@ -1627,17 +1634,15 @@ static Linehist *linehist;
 static void
 checknesting(void)
 {
-	int i;
-
 	if (includetop < 0) {
-		diag("corrupt z stack");
+		diag("dwarf: corrupt z stack");
 		errorexit();
 	}
-	if (includetop >= nelem(includestack)) {
-		diag("nesting too deep");
-		for (i = 0; i < nelem(includestack); i++)
-			diag("\t%s", histfile[includestack[i].file]);
-		errorexit();
+	if (includetop >= includestacksize) {
+		includestacksize += 1;
+		includestacksize <<= 2;
+//		print("checknesting: growing to %d\n", includestacksize);
+		includestack = realloc(includestack, includestacksize * sizeof *includestack);	       
 	}
 }
 
@@ -1660,13 +1665,14 @@ inithist(Auto *a)
 	// We have a new history.  They are guaranteed to come completely
 	// at the beginning of the compilation unit.
 	if (a->aoffset != 1) {
-		diag("stray 'z' with offset %d", a->aoffset);
+		diag("dwarf: stray 'z' with offset %d", a->aoffset);
 		return 0;
 	}
 
 	// Clear the history.
 	clearhistfile();
 	includetop = 0;
+	checknesting();
 	includestack[includetop].file = 0;
 	includestack[includetop].line = -1;
 	absline = 0;
@@ -1680,10 +1686,10 @@ inithist(Auto *a)
 	for (; a; a = a->link) {
 		if (a->type == D_FILE) {  // 'z'
 			int f = addhistfile(a->asym->name);
-			if (f < 0) {	   // pop file
+			if (f < 0) {	// pop file
 				includetop--;
 				checknesting();
-			} else if(f != includestack[includetop].file) { // pushed a new file
+			} else {	// pushed a file (potentially same)
 				includestack[includetop].line += a->aoffset - absline;
 				includetop++;
 				checknesting();
@@ -1695,7 +1701,7 @@ inithist(Auto *a)
 			// We could just fixup the current
 			// linehist->line, but there doesn't appear to
 			// be a guarantee that every 'Z' is preceded
-			// by it's own 'z', so do the safe thing and
+			// by its own 'z', so do the safe thing and
 			// update the stack and push a new Linehist
 			// entry
 			includestack[includetop].line =	 a->aoffset;
@@ -1801,7 +1807,7 @@ mkvarname(char* name, int da)
 
 // flush previous compilation unit.
 static void
-flushunit(DWDie *dwinfo, vlong pc, vlong unitstart)
+flushunit(DWDie *dwinfo, vlong pc, vlong unitstart, int32 header_length)
 {
 	vlong here;
 
@@ -1813,13 +1819,13 @@ flushunit(DWDie *dwinfo, vlong pc, vlong unitstart)
 		cput(0);  // start extended opcode
 		uleb128put(1);
 		cput(DW_LNE_end_sequence);
-		cflush();
 
 		here = cpos();
-		seek(cout, unitstart, 0);
-		LPUT(here - unitstart - sizeof(int32));
-		cflush();
-		seek(cout, here, 0);
+		cseek(unitstart);
+		LPUT(here - unitstart - sizeof(int32));	 // unit_length
+		WPUT(3);  // dwarf version
+		LPUT(header_length); // header length starting here
+		cseek(here);
 	}
 }
 
@@ -1829,7 +1835,7 @@ writelines(void)
 	Prog *q;
 	Sym *s;
 	Auto *a;
-	vlong unitstart, offs;
+	vlong unitstart, headerend, offs;
 	vlong pc, epc, lc, llc, lline;
 	int currfile;
 	int i, lang, da, dt;
@@ -1839,7 +1845,9 @@ writelines(void)
 	char *n, *nn;
 
 	unitstart = -1;
-	epc = pc = 0;
+	headerend = -1;
+	pc = 0;
+	epc = 0;
 	lc = 1;
 	llc = 1;
 	currfile = -1;
@@ -1855,7 +1863,7 @@ writelines(void)
 		// we're entering a new compilation unit
 
 		if (inithist(s->autom)) {
-			flushunit(dwinfo, epc, unitstart);
+			flushunit(dwinfo, epc, unitstart, headerend - unitstart - 10);
 			unitstart = cpos();
 
 			if(debug['v'] > 1) {
@@ -1876,10 +1884,10 @@ writelines(void)
 
 			// Write .debug_line Line Number Program Header (sec 6.2.4)
 			// Fields marked with (*) must be changed for 64-bit dwarf
-			LPUT(0);   // unit_length (*), will be filled in later.
+			LPUT(0);   // unit_length (*), will be filled in by flushunit.
 			WPUT(3);   // dwarf version (appendix F)
-			LPUT(11);  // header_length (*), starting here.
-
+			LPUT(0);   // header_length (*), filled in by flushunit.
+			// cpos == unitstart + 4 + 2 + 4
 			cput(1);   // minimum_instruction_length
 			cput(1);   // default_is_stmt
 			cput(LINE_BASE);     // line_base
@@ -1890,18 +1898,17 @@ writelines(void)
 			cput(1);   // standard_opcode_lengths[3]
 			cput(1);   // standard_opcode_lengths[4]
 			cput(0);   // include_directories  (empty)
-			cput(0);   // file_names (empty) (emitted by DW_LNE's below)
-			// header_length ends here.
 
 			for (i=1; i < histfilesize; i++) {
-				cput(0);  // start extended opcode
-				uleb128put(1 + strlen(histfile[i]) + 4);
-				cput(DW_LNE_define_file);
 				strnput(histfile[i], strlen(histfile[i]) + 4);
 				// 4 zeros: the string termination + 3 fields.
 			}
 
-			epc = pc = s->text->pc;
+			cput(0);   // terminate file_names.
+			headerend = cpos();
+
+			pc = s->text->pc;
+			epc = pc;
 			currfile = 1;
 			lc = 1;
 			llc = 1;
@@ -1915,7 +1922,7 @@ writelines(void)
 			continue;
 
 		if (unitstart < 0) {
-			diag("reachable code before seeing any history: %P", s->text);
+			diag("dwarf: reachable code before seeing any history: %P", s->text);
 			continue;
 		}
 
@@ -1932,7 +1939,7 @@ writelines(void)
 		for(q = s->text; q != P; q = q->link) {
 			lh = searchhist(q->line);
 			if (lh == nil) {
-				diag("corrupt history or bad absolute line: %P", q);
+				diag("dwarf: corrupt history or bad absolute line: %P", q);
 				continue;
 			}
 
@@ -1990,7 +1997,7 @@ writelines(void)
 			newrefattr(dwvar, DW_AT_type, defgotype(a->gotype));
 
 			// push dwvar down dwfunc->child to preserve order
-			newattr(dwvar, DW_AT_internal_location, DW_CLS_CONSTANT, offs, NULL);
+			newattr(dwvar, DW_AT_internal_location, DW_CLS_CONSTANT, offs, nil);
 			dwfunc->child = dwvar->link;  // take dwvar out from the top of the list
 			for (dws = &dwfunc->child; *dws != nil; dws = &(*dws)->link)
 				if (offs > getattr(*dws, DW_AT_internal_location)->value)
@@ -2004,7 +2011,7 @@ writelines(void)
 		dwfunc->hash = nil;
 	}
 
-	flushunit(dwinfo, epc, unitstart);
+	flushunit(dwinfo, epc, unitstart, headerend - unitstart - 10);
 	linesize = cpos() - lineo;
 }
 
@@ -2066,7 +2073,7 @@ writeframes(void)
 	// 4 is to exclude the length field.
 	pad = CIERESERVE + frameo + 4 - cpos();
 	if (pad < 0) {
-		diag("CIERESERVE too small by %lld bytes.", -pad);
+		diag("dwarf: CIERESERVE too small by %lld bytes.", -pad);
 		errorexit();
 	}
 	strnput("", pad);
@@ -2099,17 +2106,14 @@ writeframes(void)
 		pad = rnd(fdesize, PtrSize) - fdesize;
 		strnput("", pad);
 		fdesize += pad;
-		cflush();
 
 		// Emit the FDE header for real, Section 6.4.1.
-		seek(cout, fdeo, 0);
+		cseek(fdeo);
 		LPUT(fdesize);
 		LPUT(0);
 		addrput(p->pc);
 		addrput(s->size);
-
-		cflush();
-		seek(cout, fdeo + 4 + fdesize, 0);
+		cseek(fdeo + 4 + fdesize);
 	}
 
 	cflush();
@@ -2145,14 +2149,12 @@ writeinfo(void)
 
 		putdie(compunit);
 
-		cflush();
 		here = cpos();
-		seek(cout, unitstart, 0);
+		cseek(unitstart);
 		LPUT(here - unitstart - 4);	// exclude the length field.
-		cflush();
-		seek(cout, here, 0);
+		cseek(here);
 	}
-
+	cflush();
 }
 
 /*
@@ -2207,12 +2209,10 @@ writepub(int (*ispub)(DWDie*))
 		}
 		LPUT(0);
 
-		cflush();
 		here = cpos();
-		seek(cout, sectionstart, 0);
+		cseek(sectionstart);
 		LPUT(here - sectionstart - 4);	// exclude the length field.
-		cflush();
-		seek(cout, here, 0);
+		cseek(here);
 
 	}
 
@@ -2296,6 +2296,9 @@ dwarfemitdebugsections(void)
 	vlong infoe;
 	DWDie* die;
 
+	if(debug['w'])  // disable dwarf
+		return;
+
 	// For diagnostic messages.
 	newattr(&dwtypes, DW_AT_name, DW_CLS_STRING, strlen("dwtypes"), "dwtypes");
 
@@ -2340,19 +2343,23 @@ dwarfemitdebugsections(void)
 
 	infoo = cpos();
 	writeinfo();
-	gdbscripto = arangeso = pubtypeso = pubnameso = infoe = cpos();
+	infoe = cpos();
+	pubnameso = infoe;
+	pubtypeso = infoe;
+	arangeso = infoe;
+	gdbscripto = infoe;
 
 	if (fwdcount > 0) {
 		if (debug['v'])
 			Bprint(&bso, "%5.2f dwarf pass 2.\n", cputime());
-		seek(cout, infoo, 0);
+		cseek(infoo);
 		writeinfo();
 		if (fwdcount > 0) {
-			diag("unresolved references after first dwarf info pass");
+			diag("dwarf: unresolved references after first dwarf info pass");
 			errorexit();
 		}
 		if (infoe != cpos()) {
-			diag("inconsistent second dwarf info pass");
+			diag("dwarf: inconsistent second dwarf info pass");
 			errorexit();
 		}
 	}
@@ -2401,6 +2408,9 @@ vlong elfstrdbg[NElfStrDbg];
 void
 dwarfaddshstrings(Sym *shstrtab)
 {
+	if(debug['w'])  // disable dwarf
+		return;
+
 	elfstrdbg[ElfStrDebugAbbrev]   = addstring(shstrtab, ".debug_abbrev");
 	elfstrdbg[ElfStrDebugAranges]  = addstring(shstrtab, ".debug_aranges");
 	elfstrdbg[ElfStrDebugFrame]    = addstring(shstrtab, ".debug_frame");
@@ -2420,6 +2430,9 @@ dwarfaddelfheaders(void)
 {
 	ElfShdr *sh;
 
+	if(debug['w'])  // disable dwarf
+		return;
+
 	sh = newElfShdr(elfstrdbg[ElfStrDebugAbbrev]);
 	sh->type = SHT_PROGBITS;
 	sh->off = abbrevo;
@@ -2488,6 +2501,9 @@ dwarfaddmachoheaders(void)
 	vlong fakestart;
 	int nsect;
 
+	if(debug['w'])  // disable dwarf
+		return;
+
 	// Zero vsize segments won't be loaded in memory, even so they
 	// have to be page aligned in the file.
 	fakestart = abbrevo & ~0xfff;
@@ -2562,7 +2578,9 @@ dwarfaddmachoheaders(void)
 void
 dwarfaddpeheaders(void)
 {
-	dwarfemitdebugsections();
+	if(debug['w'])  // disable dwarf
+		return;
+
 	newPEDWARFSection(".debug_abbrev", abbrevsize);
 	newPEDWARFSection(".debug_line", linesize);
 	newPEDWARFSection(".debug_frame", framesize);
