@@ -328,15 +328,16 @@ ldelf(Biobuf *f, char *pkg, int64 len, char *pn)
 	Reloc *r, *rp;
 	Sym *s;
 
+	USED(pkg);
 	if(debug['v'])
 		Bprint(&bso, "%5.2f ldelf %s\n", cputime(), pn);
 
 	version++;
 	base = Boffset(f);
 
-	if(Bread(f, &hdrbuf, sizeof hdrbuf) != sizeof hdrbuf)
+	if(Bread(f, hdrbuf, sizeof hdrbuf) != sizeof hdrbuf)
 		goto bad;
-	hdr = (ElfHdrBytes*)&hdrbuf;
+	hdr = (ElfHdrBytes*)hdrbuf;
 	if(memcmp(hdr->ident, ElfMagic, 4) != 0)
 		goto bad;
 	switch(hdr->ident[5]) {
@@ -397,13 +398,13 @@ ldelf(Biobuf *f, char *pkg, int64 len, char *pn)
 		goto bad;
 
 	if(e->e16(hdr->type) != ElfTypeRelocatable) {
-		diag("%s: elf but not elf relocatable object");
+		diag("%s: elf but not elf relocatable object", pn);
 		return;
 	}
 
 	switch(thechar) {
 	default:
-		diag("%s: elf %s unimplemented", thestring);
+		diag("%s: elf %s unimplemented", pn, thestring);
 		return;
 	case '5':
 		if(e != &le || obj->machine != ElfMachArm || hdr->ident[4] != ElfClass32) {
@@ -518,7 +519,7 @@ ldelf(Biobuf *f, char *pkg, int64 len, char *pn)
 		name = smprint("%s(%s)", pn, sect->name);
 		s = lookup(name, version);
 		free(name);
-		switch(sect->flags&(ElfSectFlagAlloc|ElfSectFlagWrite|ElfSectFlagExec)) {
+		switch((int)sect->flags&(ElfSectFlagAlloc|ElfSectFlagWrite|ElfSectFlagExec)) {
 		default:
 			werrstr("unexpected flags for ELF section %s", sect->name);
 			goto bad;
