@@ -10,7 +10,6 @@ import (
 	"testing"
 )
 
-
 var tests = []struct {
 	x interface{} // x is printed as s
 	s string
@@ -49,11 +48,10 @@ var tests = []struct {
 		3  }`},
 }
 
-
 // Split s into lines, trim whitespace from all lines, and return
 // the concatenated non-empty lines.
 func trim(s string) string {
-	lines := strings.Split(s, "\n", -1)
+	lines := strings.Split(s, "\n")
 	i := 0
 	for _, line := range lines {
 		line = strings.TrimSpace(line)
@@ -65,7 +63,6 @@ func trim(s string) string {
 	return strings.Join(lines[0:i], "\n")
 }
 
-
 func TestPrint(t *testing.T) {
 	var buf bytes.Buffer
 	for _, test := range tests {
