@@ -4,14 +4,11 @@
 
 package os
 
-func (file *File) Readdirnames(count int) (names []string, err Error) {
-	fis, e := file.Readdir(count)
-	if e != nil {
-		return nil, e
-	}
+func (file *File) Readdirnames(n int) (names []string, err Error) {
+	fis, err := file.Readdir(n)
 	names = make([]string, len(fis))
 	for i, fi := range fis {
 		names[i] = fi.Name
 	}
-	return names, nil
+	return names, err
 }
