@@ -20,11 +20,16 @@ libcgo_sys_thread_start(ThreadStart *ts)
 	pthread_attr_t attr;
 	pthread_t p;
 	size_t size;
+	int err;
 
 	pthread_attr_init(&attr);
 	pthread_attr_getstacksize(&attr, &size);
 	ts->g->stackguard = size;
-	pthread_create(&p, &attr, threadentry, ts);
+	err = pthread_create(&p, &attr, threadentry, ts);
+	if (err != 0) {
+		fprintf(stderr, "runtime/cgo: pthread_create failed: %s\n", strerror(err));
+		abort();
+	}
 }
 
 static void*
