@@ -8,6 +8,7 @@
 
 extern byte end[];
 static byte *bloc = { end };
+static Lock memlock;
 
 enum
 {
@@ -19,23 +20,31 @@ runtime·SysAlloc(uintptr nbytes)
 {
 	uintptr bl;
 	
+	runtime·lock(&memlock);
+	mstats.sys += nbytes;
 	// Plan 9 sbrk from /sys/src/libc/9sys/sbrk.c
 	bl = ((uintptr)bloc + Round) & ~Round;
-	if(runtime·brk_((void*)(bl + nbytes)) < 0)
+	if(runtime·brk_((void*)(bl + nbytes)) < 0) {
+		runtime·unlock(&memlock);
 		return (void*)-1;
+	}
 	bloc = (byte*)bl + nbytes;
+	runtime·unlock(&memlock);
 	return (void*)bl;
 }
 
 void
 runtime·SysFree(void *v, uintptr nbytes)
 {
+	runtime·lock(&memlock);
+	mstats.sys -= nbytes;
 	// from tiny/mem.c
 	// Push pointer back if this is a free
 	// of the most recent SysAlloc.
 	nbytes += (nbytes + Round) & ~Round;
 	if(bloc == (byte*)v+nbytes)
 		bloc -= nbytes;	
+	runtime·unlock(&memlock);
 }
 
 void
